(* ::Package:: *)

(* Mathematica code for the evolutoion of shperically
symmetric massless scalar field coupled to gravity
with negative cosmological constant.

2013, M.Maliborski, A.Rostworowski 

Reference:
Maciej Maliborski and Andrzej Rostworowski,
"Lecture Notes on Turbulent Instability of Anti-de Sitter Spacetime" *)


(* ::Subsubsection:: *)
(* Tools *)


(* Derivatives operators *)

FDAder[o_,{x_,f_}] :=
    Module[ {},
        NDSolve`FiniteDifferenceDerivative[Derivative[o],x,f,"DifferenceOrder"->4]
    ];

FDAder[o_,{x_,f_},"Even"] :=
    Module[ {X,F},
        X = Flatten@{-x[[3]],-x[[2]],x};
        F = Flatten@{f[[3]],f[[2]],f};
        NDSolve`FiniteDifferenceDerivative[Derivative[o],X,F,"DifferenceOrder"->4][[3;;]]
    ];
    
FDAder[o_,{x_,f_},"Odd"] :=
    Module[ {X,F},
        X = Flatten@{-x[[3]],-x[[2]],x};
        F = Flatten@{-f[[3]],-f[[2]],f};
        NDSolve`FiniteDifferenceDerivative[Derivative[o],X,F,"DifferenceOrder"->4][[3;;]]
    ];
    
(* Dissipation operator *)

FDAQD[r_,{x_,f_},"Even"] :=
    Module[ {X,F,h=x[[2]]-x[[1]]},
        X = Flatten@{-x[[3]],-x[[2]],x};
        F = Flatten@{f[[3]],f[[2]],f};
        (-1/4)^r 1/h NDSolve`FiniteDifferenceDerivative[Derivative[2 r],Range[Length[X]],F,"DifferenceOrder"->2][[3;;]]
    ];

FDAQD[r_,{x_,f_},"Odd"] :=
    Module[ {X,F,h=x[[2]]-x[[1]]},
        X = Flatten@{-x[[3]],-x[[2]],x};
        F = Flatten@{-f[[3]],-f[[2]],f};
        (-1/4)^r 1/h NDSolve`FiniteDifferenceDerivative[Derivative[2 r],Range[Length[X]],F,"DifferenceOrder"->2][[3;;]]
    ];
    
(* Additional functions *)

Interior[l_] :=
    Take[l,{2,-2}]
Interpolate[f_,"Even"] :=
    1/10 (15 f[[1]]-6 f[[2]]+f[[3]])


(* ::Subsubsection:: *)
(* Constraints solver *)


Constraints[{phi_List,pi_List}] :=
    Module[ {\[CapitalPhi],\[CapitalPi],A,\[Delta],eq,g = 0,csol,\[Epsilon] = 10^-6,x},
        {\[CapitalPhi],\[CapitalPi]} = Interpolation[Transpose@{\[FormalCapitalX],#}, Method->"Spline", InterpolationOrder->4]&/@{phi,pi};
        eq = {A'[x]==(1+2Sin[x]^2)/(Sin[x]Cos[x])(1-A[x])-Sin[x]Cos[x]A[x](\[CapitalPhi][x]^2+\[CapitalPi][x]^2),
            \[Delta]'[x]==-Sin[x]Cos[x](\[CapitalPhi][x]^2+\[CapitalPi][x]^2),
            A[\[Epsilon]]==1-1/3 \[CapitalPi][0]^2 \[Epsilon]^2,
            \[Delta][\[Epsilon]]==g-1/2 \[CapitalPi][0]^2 \[Epsilon]^2};
        csol = First@NDSolve[eq,{A,\[Delta]},{x,\[Epsilon],\[Pi]/2},
            Method->{
                "ExplicitRungeKutta",
                "DifferenceOrder"->4,
                "StiffnessTest"->False
                },
            MaxSteps->Infinity];
        {Flatten@{1,A[Rest[\[FormalCapitalX]]]/.csol},Flatten@{g,\[Delta][Rest[\[FormalCapitalX]]]/.csol}}
    ]/;Length[phi]==Length[pi]==Length[\[FormalCapitalX]];


(* ::Subsubsection:: *)
(* Velocities solver *)


Velocities[{phi_List, pi_List}, opts__] :=
    Velocities[Flatten[{Constraints[{phi, pi}], {phi, pi}}, 1], opts];

Velocities[{a_List,delta_List,phi_List,pi_List},{lr_,rr_,eps_}] :=
    Module[ {d1,d2,d3,d4,d\[CapitalPi],d\[CapitalPhi],t1,t2,r,s},
    (*Left and right indices for l'Hopitals rule*)
        {r, s} = Round[(Length[\[FormalCapitalX]]-1)*{lr, 1-rr}];
    (*Calculating spatial derivatives*)
        d1 = FDAder[1,{\[FormalCapitalX],a Exp[-delta]phi},"Odd"];
        d2 = FDAder[1,{\[FormalCapitalX],a Exp[-delta]pi},"Even"];
        d3 = FDAder[1,{\[FormalCapitalX],Tan[\[FormalCapitalX]]^2 a Exp[-delta]phi},"Odd"];
        d4 = FDAder[1,{\[FormalCapitalX],Flatten[{Interpolate[#,"Even"],#}]},"Even"]&@(Rest[a Exp[-delta]phi]/(Sin[Rest[\[FormalCapitalX]]]Cos[Rest[\[FormalCapitalX]]]));
        (*Terms for l'Hopitals rule*)
        t1 = Interior[d3]/Tan[Interior[\[FormalCapitalX]]]^2;
        t2 = (1+2/Cos[2 Interior[\[FormalCapitalX]]])Interior[d1]-Tan[2 Interior[\[FormalCapitalX]]]*Interior[d4];
        (*Time derivatives*)
        d\[CapitalPi] = Flatten[{t2[[;;r]],t1[[r+1;;s-1]],t2[[s;;]]}];
        d\[CapitalPhi] = Interior[d2];
        (*Adding boundary points*)
        {d\[CapitalPhi],d\[CapitalPi]} = Flatten/@{{0,d\[CapitalPhi],0},{Interpolate[d\[CapitalPi],"Even"],d\[CapitalPi],0}};
        (*Adding artificial dissipation*)
        {d\[CapitalPhi]-eps Flatten@{Most@FDAQD[3,{\[FormalCapitalX],phi},"Odd"],0},d\[CapitalPi]-eps Flatten@{Most@FDAQD[3,{\[FormalCapitalX],pi},"Even"],0}}
    ];



(* ::Subsubsection:: *)
(* Dynamical solver *)


Options[Solver] = {"DissipationStrength"->1/10,"lHopitalsRule"->{1/4, 1/4}};

Solver::npts = "Number of grid points must be an odd integer."
Solver::diss = "Dissipation strength parameter `1` must be a small positive number.";
Solver::lhsp = "Parameter `1` for l'Hopital's rule is not two element list.";
Solver::lhsl = "Parameter `1` for l'Hopital's rule is not properly defined interval with lower and upper bounds \[Element] [0,1/4]";

Solver[{x_,{phi_,pi_}},{t0_,t1_},opts:OptionsPattern[]] :=
    Module[ {eps,lh},
        If[ Not[OddQ[Length[x]]],
            Message[Solver::npts];
            Abort[]
        ];
        {eps,lh} = OptionValue[Solver,FilterRules[{opts},Options[Solver]],#]&/@{"DissipationStrength","lHopitalsRule"};
        (*Check input options*)
        If[ Not[0<=eps<1],
            Message[Solver::diss,eps];
            Abort[]
        ];
        If[ Not@MatchQ[lh,{_?NumericQ,_?NumericQ}],
            Message[Solver::lhsp,lh];
            Abort[]
        ];
        If[ Not[And[0<=lh[[1]]<=1/4, 0<=lh[[2]]<=1/4]],
            Message[Solver::lhsl,lh];
            Abort[]
        ];
        (*Integrate equations*)
        Block[ {\[FormalCapitalX] = x},
            NDSolve[{u'[t]==Velocities[u[t],{lh[[1]],lh[[2]],eps}],u[t0]=={phi,pi}},u,{t,t0,t1},FilterRules[{opts},Options[NDSolve]]]
        ]
    ]/;Length[x]==Length[phi]==Length[pi];
